@@ -12,7 +12,6 @@ import (
 	"crypto/rsa"
 	"crypto/subtle"
 	"crypto/x509"
-	"encoding/binary"
 	"errors"
 	"fmt"
 	"io"
@@ -20,8 +19,12 @@ import (
 	"strings"
 	"sync/atomic"
 	"time"
+
+	"golang.org/x/crypto/cryptobyte"
 )
 
+const clientSessionStateVersion = 1
+
 type clientHandshakeState struct {
 	c            *Conn
 	serverHello  *serverHelloMsg
@@ -67,7 +70,6 @@ func (c *Conn) makeClientHello() (*clientHelloMsg, ecdheParameters, error) {
 		vers:                         clientHelloVersion,
 		compressionMethods:           []uint8{compressionNone},
 		random:                       make([]byte, 32),
-		sessionId:                    make([]byte, 32),
 		ocspStapling:                 true,
 		scts:                         true,
 		serverName:                   hostnameInSNI(config.ServerName),
@@ -111,8 +113,11 @@ func (c *Conn) makeClientHello() (*clientHelloMsg, ecdheParameters, error) {
 	// A random session ID is used to detect when the server accepted a ticket
 	// and is resuming a session (see RFC 5077). In TLS 1.3, it's always set as
 	// a compatibility measure (see RFC 8446, Section 4.1.2).
-	if _, err := io.ReadFull(config.rand(), hello.sessionId); err != nil {
-		return nil, nil, errors.New("tls: short read from Rand: " + err.Error())
+	if c.config.AlternativeRecordLayer == nil {
+		hello.sessionId = make([]byte, 32)
+		if _, err := io.ReadFull(config.rand(), hello.sessionId); err != nil {
+			return nil, nil, errors.New("tls: short read from Rand: " + err.Error())
+		}
 	}
 
 	if hello.vers >= VersionTLS12 {
@@ -256,6 +261,33 @@ func (c *Conn) clientHandshake() (err error) {
 	return nil
 }
 
+// extract the app data saved in the session.nonce,
+// and set the session.nonce to the actual nonce value
+func (c *Conn) decodeSessionState(session *ClientSessionState) (uint32 /* max early data */, []byte /* app data */, bool /* ok */) {
+	s := cryptobyte.String(session.nonce)
+	var version uint16
+	if !s.ReadUint16(&version) {
+		return 0, nil, false
+	}
+	if version != clientSessionStateVersion {
+		return 0, nil, false
+	}
+	var maxEarlyData uint32
+	if !s.ReadUint32(&maxEarlyData) {
+		return 0, nil, false
+	}
+	var appData []byte
+	if !readUint16LengthPrefixed(&s, &appData) {
+		return 0, nil, false
+	}
+	var nonce []byte
+	if !readUint16LengthPrefixed(&s, &nonce) {
+		return 0, nil, false
+	}
+	session.nonce = nonce
+	return maxEarlyData, appData, true
+}
+
 func (c *Conn) loadSession(hello *clientHelloMsg) (cacheKey string,
 	session *ClientSessionState, earlySecret, binderKey []byte) {
 	if c.config.SessionTicketsDisabled || c.config.ClientSessionCache == nil {
@@ -283,6 +315,16 @@ func (c *Conn) loadSession(hello *clientHelloMsg) (cacheKey string,
 	if !ok || session == nil {
 		return cacheKey, nil, nil, nil
 	}
+	var appData []byte
+	var maxEarlyData uint32
+	if session.vers == VersionTLS13 {
+		var ok bool
+		maxEarlyData, appData, ok = c.decodeSessionState(session)
+		if !ok { // delete it, if parsing failed
+			c.config.ClientSessionCache.Put(cacheKey, nil)
+			return cacheKey, nil, nil, nil
+		}
+	}
 
 	// Check that version used for the previous session is still valid.
 	versOk := false
@@ -326,14 +368,6 @@ func (c *Conn) loadSession(hello *clientHelloMsg) (cacheKey string,
 		return
 	}
 
-	// In TLS 1.3, we abuse the nonce field to save the max_early_data_size.
-	// See Conn.handleNewSessionTicket for an explanation of this hack.
-	if len(session.nonce) < 4 {
-		return cacheKey, nil, nil, nil
-	}
-	maxEarlyData := binary.BigEndian.Uint32(session.nonce[:4])
-	session.nonce = session.nonce[4:]
-
 	// Check that the session ticket is not expired.
 	if c.config.time().After(session.useBy) {
 		c.config.ClientSessionCache.Put(cacheKey, nil)
@@ -378,6 +412,9 @@ func (c *Conn) loadSession(hello *clientHelloMsg) (cacheKey string,
 	pskBinders := [][]byte{cipherSuite.finishedHash(binderKey, transcript)}
 	hello.updateBinders(pskBinders)
 
+	if session.vers == VersionTLS13 && c.config.SetAppDataFromSessionState != nil {
+		c.config.SetAppDataFromSessionState(appData)
+	}
 	return
 }
 
