@@ -599,7 +599,7 @@ func TestWarningAlertFlood(t *testing.T) {
 }
 
 func TestCloneFuncFields(t *testing.T) {
-	const expectedCount = 9
+	const expectedCount = 11
 	called := 0
 
 	c1 := Config{
@@ -637,6 +637,13 @@ func TestCloneFuncFields(t *testing.T) {
 		Rejected0RTT: func() {
 			called |= 1 << 8
 		},
+		GetAppDataForSessionState: func() []byte {
+			called |= 1 << 9
+			return nil
+		},
+		SetAppDataFromSessionState: func([]byte) {
+			called |= 1 << 10
+		},
 	}
 
 	c2 := c1.Clone()
@@ -650,6 +657,8 @@ func TestCloneFuncFields(t *testing.T) {
 	c2.ReceivedExtensions(0, nil)
 	c2.Accept0RTT(nil)
 	c2.Rejected0RTT()
+	c2.GetAppDataForSessionState()
+	c2.SetAppDataFromSessionState(nil)
 
 	if called != (1<<expectedCount)-1 {
 		t.Fatalf("expected %d calls but saw calls %b", expectedCount, called)
@@ -673,7 +682,7 @@ func TestCloneNonFuncFields(t *testing.T) {
 		switch fn := typ.Field(i).Name; fn {
 		case "Rand":
 			f.Set(reflect.ValueOf(io.Reader(os.Stdin)))
-		case "Time", "GetCertificate", "GetConfigForClient", "VerifyPeerCertificate", "GetClientCertificate", "GetExtensions", "ReceivedExtensions", "Accept0RTT", "Rejected0RTT":
+		case "Time", "GetCertificate", "GetConfigForClient", "VerifyPeerCertificate", "GetClientCertificate", "GetExtensions", "ReceivedExtensions", "Accept0RTT", "Rejected0RTT", "GetAppDataForSessionState", "SetAppDataFromSessionState":
 			// DeepEqual can't compare functions. If you add a
 			// function field to this list, you must also change
 			// TestCloneFuncFields to ensure that the func field is
